import { Mesh, ConeGeometry, MeshBasicMaterial, LineLoop, Vector3, Group, BufferGeometry, Float32BufferAttribute, OctahedronBufferGeometry, DoubleSide, PlaneGeometry, Quaternion, MathUtils, Raycaster as Raycaster$1, Vector2, Plane as Plane$1, PlaneHelper, Object3D } from 'three';

const DEFAULT_LINE_HEIGHT = 1;
const DEFAULT_RADIAL_SEGMENTS = 32;
const DEFAULT_CONE_HEIGHT = 0.75;
const DEFAULT_CONE_RADIUS = 0.3;
const DEFAULT_OCTAHEDRON_RADIUS = 0.1;
const DEFAULT_PLANE_WIDTH = 0.75;
const DEFAULT_PLANE_HEIGHT = 0.75;
const DEFAULT_PLANE_SEGMENTS = 32;
const DEFAULT_RING_NUM_POINTS = 64;
const DEFAULT_RING_RADIUS = 1;
var PICK_PLANE_OPACITY;
(function (PICK_PLANE_OPACITY) {
    PICK_PLANE_OPACITY[PICK_PLANE_OPACITY["ACTIVE"] = 0.75] = "ACTIVE";
    PICK_PLANE_OPACITY[PICK_PLANE_OPACITY["INACTIVE"] = 0.3] = "INACTIVE";
})(PICK_PLANE_OPACITY || (PICK_PLANE_OPACITY = {}));
const DEFAULT_CONTROLS_SEPARATION = 1;
const DEFAULT_ROTATION_RADIUS_SCALE = 1;
const DEFAULT_EYE_ROTATION_SCALE = 1.25;
const DEFAULT_PLANE_SIZE_SCALE = 0.75;
const DEFAULT_TRANSLATION_DISTANCE_SCALE = 1;
const DEFAULT_COLOR_ARROW = "#f0ff00";
const DEFAULT_COLOR_RING = "#f0ff00";
const DEFAULT_COLOR_PLANE = "#f0ff00";
const DEFAULT_CONTROLS_OPACITY = 1;

class Cone extends Mesh {
    constructor(color) {
        super();
        this.geometry = new ConeGeometry(DEFAULT_CONE_RADIUS, DEFAULT_CONE_HEIGHT, DEFAULT_RADIAL_SEGMENTS);
        this.material = new MeshBasicMaterial({ color, depthTest: false });
        this.material.transparent = true;
        this.material.opacity = DEFAULT_CONTROLS_OPACITY;
    }
}

class Line extends LineLoop {
    constructor(color, geometry) {
        super();
        this.geometry = geometry;
        this.material = new MeshBasicMaterial({ color, depthTest: true });
        this.material.transparent = true;
        this.material.opacity = DEFAULT_CONTROLS_OPACITY;
    }
}

/**
 * Names for default handles
 */
var DEFAULT_HANDLE_GROUP_NAME;
(function (DEFAULT_HANDLE_GROUP_NAME) {
    /**
     * name for default translation handle along the +ve x-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["XPT"] = "xpt_handle";
    /**
     * name for default translation handle along the +ve y-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["YPT"] = "ypt_handle";
    /**
     * name for default translation handle along the +ve z-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["ZPT"] = "zpt_handle";
    /**
     * name for default translation handle along the -ve x-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["XNT"] = "xnt_handle";
    /**
     * name for default translation handle along the -ve y-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["YNT"] = "ynt_handle";
    /**
     * name for default translation handle along the -ve z-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["ZNT"] = "znt_handle";
    /**
     * name for default rotation handle along the x-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["XR"] = "xr_handle";
    /**
     * name for default rotation handle along the y-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["YR"] = "yr_handle";
    /**
     * name for default rotation handle along the z-axis
     */
    DEFAULT_HANDLE_GROUP_NAME["ZR"] = "zr_handle";
    /**
     * name for default rotation handle in the eye-plane
     */
    DEFAULT_HANDLE_GROUP_NAME["ER"] = "er_handle";
    /**
     * name for default translation handle in the eye-plane
     */
    DEFAULT_HANDLE_GROUP_NAME["PICK"] = "pick_handle";
    /**
     * name for default translation handle in the xy plane
     */
    DEFAULT_HANDLE_GROUP_NAME["PICK_PLANE_XY"] = "pick_plane_xy_handle";
    /**
     * name for default translation handle in the yz plane
     */
    DEFAULT_HANDLE_GROUP_NAME["PICK_PLANE_YZ"] = "pick_plane_yz_handle";
    /**
     * name for default translation handle in the zx plane
     */
    DEFAULT_HANDLE_GROUP_NAME["PICK_PLANE_ZX"] = "pick_plane_zx_handle";
})(DEFAULT_HANDLE_GROUP_NAME || (DEFAULT_HANDLE_GROUP_NAME = {}));
/**
 * Base class for all handles
 * @noInheritDoc
 */
class HandleGroup extends Group {
}
/**
 * This class can be extended to create custom translation handles.
 * It requires setting the properties `parallel` and `up`.
 */
class TranslationGroup extends HandleGroup {
    constructor() {
        super(...arguments);
        /**
         * This is a unit vector that runs perpendicular to the direction of the translation handles.
         * For example, in case of [[Controls.translationXP]], it is
         * `THREE.Vector3(0,1,0)` (along the y-axis).
         */
        this.up = new Vector3();
    }
}
/**
 * This class can be extended to create custom rotation handles.
 * It requires setting the property `up`.
 */
class RotationGroup extends HandleGroup {
    constructor() {
        super(...arguments);
        /**
         * This is a unit vector that runs along the axis of the rotation handles.
         * For example, in case of [[Controls.rotationX]], it is
         * `THREE.Vector3(1,0,0)` (along the x-axis).
         */
        this.up = new Vector3();
    }
}
/**
 * This class can be extended to create custom pick handle (translates in eye-plane).
 */
class PickGroup extends HandleGroup {
}
/**
 * This class can be extended to create custom pick plane handles.
 * It requires setting the property `up`.
 */
class PickPlaneGroup extends HandleGroup {
    constructor() {
        super(...arguments);
        /**
         * This is a unit vector that is perpendicular to the plane handles.
         * For example, in case of [[Controls.pickPlaneXY]], it is
         * `THREE.Vector3(0,0,1)` (along the z-axis).
         */
        this.up = new Vector3();
    }
}

class Translation extends TranslationGroup {
    constructor(color = DEFAULT_COLOR_ARROW) {
        super();
        this.parallel = new Vector3(0, 1, 0);
        /**
         * @hidden
         */
        this.getInteractiveObjects = () => {
            return [this.cone];
        };
        this.setColor = (color) => {
            const coneMaterial = this.cone.material;
            const lineMaterial = this.line.material;
            coneMaterial.color.set(color);
            lineMaterial.color.set(color);
        };
        this.cone = new Cone(color);
        const lineGeometry = new BufferGeometry();
        lineGeometry.setAttribute('position', new Float32BufferAttribute([
            0, 0, 0,
            0, DEFAULT_LINE_HEIGHT, 0
        ], 3));
        this.line = new Line(color, lineGeometry);
        this.cone.geometry.scale(DEFAULT_CONE_RADIUS, DEFAULT_CONE_HEIGHT, DEFAULT_CONE_RADIUS);
        this.cone.translateY(DEFAULT_LINE_HEIGHT);
        this.add(this.cone);
        this.add(this.line);
    }
}

class Octahedron extends Mesh {
    constructor(color) {
        super();
        this.geometry = new OctahedronBufferGeometry(DEFAULT_OCTAHEDRON_RADIUS, 0);
        this.material = new MeshBasicMaterial({
            color,
            depthTest: false,
            transparent: true,
            side: DoubleSide
        });
    }
}

class Rotation extends RotationGroup {
    constructor(color = DEFAULT_COLOR_RING, ringRadius = DEFAULT_RING_RADIUS) {
        super();
        /**
         * @hidden
         */
        this.getInteractiveObjects = () => {
            return [this.handlebar];
        };
        this.setColor = (color) => {
            const ringMaterial = this.ring.material;
            const handlebarMaterial = this.handlebar.material;
            ringMaterial.color.set(color);
            handlebarMaterial.color.set(color);
        };
        const ringNumberOfPoints = DEFAULT_RING_NUM_POINTS;
        const ringGeometry = new BufferGeometry();
        const angle = (2 * Math.PI) / ringNumberOfPoints;
        const vertices = [];
        for (let i = 1; i < ringNumberOfPoints + 1; i++) {
            vertices.push(ringRadius * Math.cos(i * angle), ringRadius * Math.sin(i * angle), 0);
        }
        ringGeometry.setAttribute('position', new Float32BufferAttribute(vertices, 3));
        this.ring = new Line(color, ringGeometry);
        this.handlebar = new Octahedron(color);
        this.handlebar.position.y = ringRadius;
        this.add(this.ring);
        this.add(this.handlebar);
    }
}

class Pick extends PickGroup {
    constructor() {
        super();
        /**
         * @hidden
         */
        this.getInteractiveObjects = () => {
            return [this.octahedron];
        };
        this.setColor = (color) => {
            const octahedronMaterial = this.octahedron.material;
            octahedronMaterial.color.set(color);
        };
        this.octahedron = new Octahedron("white");
        this.add(this.octahedron);
    }
}

class Plane extends Mesh {
    constructor(color = DEFAULT_COLOR_RING, width = DEFAULT_PLANE_WIDTH, height = DEFAULT_PLANE_HEIGHT) {
        super();
        this.geometry = new PlaneGeometry(width, height, DEFAULT_PLANE_SEGMENTS);
        this.material = new MeshBasicMaterial({
            color,
            depthTest: false,
            side: DoubleSide,
            transparent: true
        });
        this.material.opacity = PICK_PLANE_OPACITY.INACTIVE;
    }
}

class PickPlane extends PickPlaneGroup {
    constructor(color = DEFAULT_COLOR_PLANE, width = DEFAULT_PLANE_WIDTH, height = DEFAULT_PLANE_HEIGHT) {
        super();
        /**
         * @hidden
         */
        this.getInteractiveObjects = () => {
            return [this.plane];
        };
        this.setColor = (color) => {
            const planeMaterial = this.plane.material;
            const boundaryMaterial = this.boundary.material;
            planeMaterial.color.set(color);
            boundaryMaterial.color.set(color);
        };
        const boundaryGeometry = new BufferGeometry();
        const crossXGeometry = new BufferGeometry();
        const crossYGeometry = new BufferGeometry();
        const vertexMaxX = width / 2;
        const vertexMaxY = height / 2;
        boundaryGeometry.setAttribute('position', new Float32BufferAttribute([
            vertexMaxX, vertexMaxY, 0,
            vertexMaxX, -vertexMaxY, 0,
            -vertexMaxX, -vertexMaxY, 0,
            -vertexMaxX, vertexMaxY, 0
        ], 3));
        crossXGeometry.setAttribute('position', new Float32BufferAttribute([
            0, vertexMaxY, 0,
            0, -vertexMaxY, 0,
        ], 3));
        crossYGeometry.setAttribute('position', new Float32BufferAttribute([
            -vertexMaxX, 0, 0,
            vertexMaxX, 0, 0,
        ], 3));
        this.boundary = new Line(color, boundaryGeometry);
        this.crossX = new Line("black", crossXGeometry);
        this.crossY = new Line("black", crossYGeometry);
        this.plane = new Plane(color, width, height);
        this.add(this.plane);
        this.add(this.boundary);
        this.add(this.crossX);
        this.add(this.crossY);
    }
}

/**
 * @hidden
 */
class RotationEye extends Rotation {
    constructor(color = DEFAULT_COLOR_RING, radius = DEFAULT_RING_RADIUS) {
        super(color, radius);
        this.camera = null;
        this.controlsWorldOrientation = new Quaternion();
        this._temp1 = new Vector3();
        this._temp2 = new Vector3();
        this._temp3 = new Quaternion();
        this.worldPosition = new Vector3();
    }
    updateMatrixWorld(force) {
        var _a;
        if (this.camera !== null) {
            (_a = this.parent) === null || _a === void 0 ? void 0 : _a.matrixWorld.decompose(this._temp1, this.controlsWorldOrientation, this._temp2);
            this.matrixWorld.decompose(this.worldPosition, this._temp3, this._temp2);
            this.camera
                .getWorldQuaternion(this.quaternion)
                .premultiply(this.controlsWorldOrientation.invert());
            this.camera.getWorldPosition(this.up).sub(this.worldPosition);
        }
        super.updateMatrixWorld(force);
    }
}

var ANCHOR_MODE;
(function (ANCHOR_MODE) {
    /**
     * In this mode the Controls do not inherit the orientation of the object
     * as it is rotated.
     */
    ANCHOR_MODE["FIXED"] = "fixed";
    /**
     * In this mode the Controls rotate as the object is rotated.
     */
    ANCHOR_MODE["INHERIT"] = "inherit";
})(ANCHOR_MODE || (ANCHOR_MODE = {}));
/**
 * Controls is the main class in this library.
 * It is a subclass of THREE.Group, so its properties like `position` and
 * `quaternion` can be modified as desired.
 * The `children` are the control handles (like `rotationX`).
 * All translations and rotations are setup with respect to the global coordinate system.
 * @noInheritDoc
 */
class Controls extends Group {
    /**
     *
     * @param object - the object provided by the user
     * @param camera - the THREE.Camera instance used in the scene
     * @param options
     */
    constructor(object, camera, options) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v, _w, _x, _y, _z, _0, _1;
        super();
        this.object = object;
        this.camera = camera;
        this.handleTargetQuaternion = new Quaternion();
        this.objectWorldPosition = new Vector3();
        this.objectTargetPosition = new Vector3();
        this.objectTargetQuaternion = new Quaternion();
        this.objectParentWorldPosition = new Vector3();
        this.objectParentWorldQuaternion = new Quaternion();
        this.objectParentWorldScale = new Vector3();
        this.deltaPosition = new Vector3();
        this.normalizedHandleParallelVectorCache = new Vector3();
        this.touch1 = new Vector3();
        this.touch2 = new Vector3();
        this.boundingSphereRadius = 0;
        this.dragStartPoint = new Vector3();
        this.dragIncrementalStartPoint = new Vector3();
        this.handles = new Set();
        this.isBeingDraggedTranslation = false;
        this.isBeingDraggedRotation = false;
        this.dampingFactor = 0.8;
        this.initialSelfQuaternion = new Quaternion();
        this.minTranslationCache = new Vector3();
        this.maxTranslationCache = new Vector3();
        this.translationLimit = false;
        this.translationAnchor = null;
        this.setupDefaultPickPlane = () => {
            this.pickPlaneXY.name = DEFAULT_HANDLE_GROUP_NAME.PICK_PLANE_XY;
            this.pickPlaneYZ.name = DEFAULT_HANDLE_GROUP_NAME.PICK_PLANE_YZ;
            this.pickPlaneZX.name = DEFAULT_HANDLE_GROUP_NAME.PICK_PLANE_ZX;
            this.pickPlaneYZ.up = new Vector3(1, 0, 0);
            this.pickPlaneZX.up = new Vector3(0, 1, 0);
            this.pickPlaneXY.up = new Vector3(0, 0, 1);
            this.pickPlaneYZ.rotateY(Math.PI / 2);
            this.pickPlaneZX.rotateX(Math.PI / 2);
            this.setupHandle(this.pickPlaneXY);
            this.setupHandle(this.pickPlaneYZ);
            this.setupHandle(this.pickPlaneZX);
        };
        this.setupHandle = (handle) => {
            this.handles.add(handle);
            this.add(handle);
        };
        this.setupDefaultPick = () => {
            this.pick.name = DEFAULT_HANDLE_GROUP_NAME.PICK;
            this.setupHandle(this.pick);
        };
        this.setupDefaultEyeRotation = () => {
            this.rotationEye.name = DEFAULT_HANDLE_GROUP_NAME.ER;
            this.rotationEye.camera = this.camera;
            this.setupHandle(this.rotationEye);
        };
        this.setupDefaultTranslation = () => {
            this.translationXP.name = DEFAULT_HANDLE_GROUP_NAME.XPT;
            this.translationYP.name = DEFAULT_HANDLE_GROUP_NAME.YPT;
            this.translationZP.name = DEFAULT_HANDLE_GROUP_NAME.ZPT;
            this.translationXN.name = DEFAULT_HANDLE_GROUP_NAME.XNT;
            this.translationYN.name = DEFAULT_HANDLE_GROUP_NAME.YNT;
            this.translationZN.name = DEFAULT_HANDLE_GROUP_NAME.ZNT;
            this.translationXP.translateX(this.boundingSphereRadius * this.translationDistanceScale);
            this.translationYP.translateY(this.boundingSphereRadius * this.translationDistanceScale);
            this.translationZP.translateZ(this.boundingSphereRadius * this.translationDistanceScale);
            this.translationXN.translateX(-this.boundingSphereRadius * this.translationDistanceScale);
            this.translationYN.translateY(-this.boundingSphereRadius * this.translationDistanceScale);
            this.translationZN.translateZ(-this.boundingSphereRadius * this.translationDistanceScale);
            this.translationXP.rotateZ(-Math.PI / 2);
            this.translationZP.rotateX(Math.PI / 2);
            this.translationXN.rotateZ(Math.PI / 2);
            this.translationYN.rotateX(Math.PI);
            this.translationZN.rotateX(-Math.PI / 2);
            this.translationXP.up = new Vector3(0, 1, 0);
            this.translationYP.up = new Vector3(0, 0, 1);
            this.translationZP.up = new Vector3(0, 1, 0);
            this.translationXN.up = new Vector3(0, -1, 0);
            this.translationYN.up = new Vector3(0, 0, -1);
            this.translationZN.up = new Vector3(0, -1, 0);
            this.translationXP.parallel = new Vector3(1, 0, 0);
            this.translationYP.parallel = new Vector3(0, 1, 0);
            this.translationZP.parallel = new Vector3(0, 0, 1);
            this.translationXN.parallel = new Vector3(-1, 0, 0);
            this.translationYN.parallel = new Vector3(0, -1, 0);
            this.translationZN.parallel = new Vector3(0, 0, -1);
            this.setupHandle(this.translationXP);
            this.setupHandle(this.translationYP);
            this.setupHandle(this.translationZP);
            this.setupHandle(this.translationXN);
            this.setupHandle(this.translationYN);
            this.setupHandle(this.translationZN);
        };
        this.setupDefaultRotation = () => {
            this.rotationX.name = DEFAULT_HANDLE_GROUP_NAME.XR;
            this.rotationY.name = DEFAULT_HANDLE_GROUP_NAME.YR;
            this.rotationZ.name = DEFAULT_HANDLE_GROUP_NAME.ZR;
            this.rotationX.up = new Vector3(1, 0, 0);
            this.rotationY.up = new Vector3(0, 1, 0);
            this.rotationZ.up = new Vector3(0, 0, 1);
            this.rotationY.rotateX(Math.PI / 2);
            this.rotationX.rotateY(Math.PI / 2);
            this.rotationX.rotateZ(Math.PI);
            this.setupHandle(this.rotationX);
            this.setupHandle(this.rotationY);
            this.setupHandle(this.rotationZ);
        };
        this.computeObjectBounds = () => {
            var _a;
            if (this.useComputedBounds) {
                if (this.object.type === "Mesh") {
                    const geometry = this.object.geometry;
                    geometry.computeBoundingSphere();
                    const { boundingSphere } = geometry;
                    const radius = (_a = boundingSphere === null || boundingSphere === void 0 ? void 0 : boundingSphere.radius) !== null && _a !== void 0 ? _a : 0;
                    this.boundingSphereRadius = radius / 2 + this.separation;
                    return;
                }
                else {
                    console.warn(`Bounds can only be computed for object of type THREE.Mesh,
          received object with type: ${this.object.type}. Falling back to using
          default separation.
        `);
                }
            }
            this.boundingSphereRadius = this.separation;
        };
        /**
         * Puts a limit on the object's translation anchored at the current position.
         *
         * `{ x: [-1, 2], y: false, z: false }` - sets the translation limit to `-1` unit
         * in the -x-direction, `+2` units in the +x-direction, and no limit on the
         * y and z-direction.
         *
         * Setting the limit to `false` disables the limit in all directions.
         * @param limit
         */
        this.setTranslationLimit = (limit) => {
            this.translationLimit = limit;
            this.translationAnchor = limit ? this.position.clone() : null;
        };
        /**
         * @hidden
         */
        this.processDragStart = (args) => {
            const { point, handle } = args;
            this.dragStartPoint.copy(point);
            this.dragIncrementalStartPoint.copy(point);
            this.isBeingDraggedTranslation =
                handle instanceof TranslationGroup ||
                    handle instanceof PickGroup ||
                    handle instanceof PickPlaneGroup;
            this.isBeingDraggedRotation = handle instanceof RotationGroup;
        };
        /**
         * @hidden
         */
        this.processDragEnd = (args) => {
            const { handle } = args;
            const { x: xSnap, y: ySnap, z: zSnap } = this.snapTranslation;
            const snap = [xSnap, ySnap, zSnap];
            if (handle instanceof TranslationGroup ||
                handle instanceof PickPlaneGroup ||
                handle instanceof PickGroup) {
                const xyz = this.object.position.toArray();
                const floor = xyz.map(Math.floor);
                const ceil = xyz.map(Math.ceil);
                const snapFloor = xyz.map((p, index) => ceil[index] - p >= p - floor[index]);
                const position = xyz.map((p, index) => {
                    if (!snap[index]) {
                        return p;
                    }
                    return snapFloor[index] ? floor[index] : ceil[index];
                });
                this.object.position.fromArray(position);
            }
            this.isBeingDraggedTranslation = false;
            this.isBeingDraggedRotation = false;
        };
        /**
         * Only takes effect if [[IControlsOptions.isDampingEnabled]] is true.
         * @param dampingFactor - value between 0 and 1, acts like a weight on the controls
         */
        this.setDampingFactor = (dampingFactor = 0) => (this.dampingFactor = MathUtils.clamp(dampingFactor, 0, 1));
        /**
         * @hidden
         */
        this.processDrag = (args) => {
            const { point, handle, dragRatio = 1 } = args;
            const k = Math.exp(-this.dampingFactor * Math.abs(Math.pow(dragRatio, 3)));
            if (handle instanceof TranslationGroup) {
                this.deltaPosition.copy(point).sub(this.dragIncrementalStartPoint);
                this.normalizedHandleParallelVectorCache
                    .copy(handle.parallel.normalize())
                    .applyQuaternion(this.quaternion);
                const delta = this.deltaPosition.dot(this.normalizedHandleParallelVectorCache);
                this.deltaPosition
                    .copy(this.normalizedHandleParallelVectorCache)
                    .multiplyScalar(this.isDampingEnabled ? k * delta : delta);
                this.position.copy(this.getLimitedTranslation(this.deltaPosition));
            }
            else if (handle instanceof PickGroup || handle instanceof PickPlaneGroup) {
                this.deltaPosition
                    .copy(point)
                    .sub(this.dragIncrementalStartPoint)
                    .multiplyScalar(this.isDampingEnabled ? k : 1);
                this.position.copy(this.getLimitedTranslation(this.deltaPosition));
            }
            else {
                this.touch1
                    .copy(this.dragIncrementalStartPoint)
                    .sub(this.objectWorldPosition)
                    .normalize();
                this.touch2
                    .copy(point)
                    .sub(this.objectWorldPosition)
                    .normalize();
                this.handleTargetQuaternion.setFromUnitVectors(this.touch1, this.touch2);
                if (this.mode === ANCHOR_MODE.FIXED) {
                    this.detachHandleUpdateQuaternionAttach(handle, this.handleTargetQuaternion);
                }
            }
            this.objectTargetQuaternion.premultiply(this.handleTargetQuaternion);
            this.dragIncrementalStartPoint.copy(point);
        };
        this.getLimitedTranslation = (translation) => {
            const position = translation.add(this.position);
            if (!this.translationAnchor || !this.translationLimit) {
                return position;
            }
            const { x: xLimit, y: yLimit, z: zLimit } = this.translationLimit;
            const { x: xAnchor, y: yAnchor, z: zAnchor } = this.translationAnchor;
            const { x, y, z } = position;
            this.minTranslationCache.set(xLimit ? xAnchor + xLimit[0] : x, yLimit ? yAnchor + yLimit[0] : y, zLimit ? zAnchor + zLimit[0] : z);
            this.maxTranslationCache.set(xLimit ? xAnchor + xLimit[1] : x, yLimit ? yAnchor + yLimit[1] : y, zLimit ? zAnchor + zLimit[1] : z);
            return position.clamp(this.minTranslationCache, this.maxTranslationCache);
        };
        this.detachObjectUpdatePositionAttach = (parent, object) => {
            if (parent !== null && this.parent !== null && this.parent.parent !== null) {
                const scene = this.parent.parent;
                if (scene.type !== "Scene") {
                    throw new Error("freeform controls must be attached to the scene");
                }
                scene.attach(object);
                object.position.copy(this.objectTargetPosition);
                parent.attach(object);
            }
        };
        this.detachHandleUpdateQuaternionAttach = (handle, quaternion) => {
            if (this.parent !== null && this.parent.parent) {
                const scene = this.parent.parent;
                if (scene.type !== "Scene") {
                    throw new Error("freeform controls must be attached to the scene");
                }
                scene.attach(handle);
                handle.applyQuaternion(quaternion);
                this.attach(handle);
            }
        };
        /**
         * Applies supplied visibility to the supplied handle names.
         * Individual handle's visibility can also be changed by modifying the `visibility`
         * property on the handle directly.
         * @param handleNames
         * @param visibility
         */
        this.showByNames = (handleNames, visibility = true) => {
            const handleNamesMap = {};
            this.handles.forEach(handle => {
                handleNamesMap[handle.name] = handle;
            });
            handleNames.map(handleName => {
                const handle = handleNamesMap[handleName];
                if (handle === undefined) {
                    throw new Error(`handle: ${handleName} not found`);
                }
                handle.visible = visibility;
            });
        };
        /**
         * Applies supplied visibility to all handles
         * @param visibility
         */
        this.showAll = (visibility = true) => {
            this.handles.forEach(handle => {
                handle.visible = visibility;
            });
        };
        /**
         * @hidden
         */
        this.updateMatrixWorld = (force) => {
            this.object.updateMatrixWorld(force);
            this.object.getWorldPosition(this.objectWorldPosition);
            const parent = this.object.parent;
            if (parent !== null) {
                parent.matrixWorld.decompose(this.objectParentWorldPosition, this.objectParentWorldQuaternion, this.objectParentWorldScale);
            }
            this.objectParentWorldQuaternion.invert();
            this.objectTargetPosition.copy(this.position);
            this.objectTargetQuaternion.premultiply(this.objectParentWorldQuaternion);
            if (this.isBeingDraggedTranslation) {
                this.detachObjectUpdatePositionAttach(parent, this.object);
            }
            else if (this.isBeingDraggedRotation) {
                this.object.quaternion.copy(this.objectTargetQuaternion);
                this.detachObjectUpdatePositionAttach(parent, this.object);
            }
            else {
                this.position.copy(this.objectWorldPosition);
            }
            this.object.getWorldQuaternion(this.objectTargetQuaternion);
            if (this.mode === ANCHOR_MODE.INHERIT && !this.isBeingDraggedTranslation) {
                this.quaternion.copy(this.initialSelfQuaternion).premultiply(this.objectTargetQuaternion);
            }
            super.updateMatrixWorld(force);
        };
        this.options = options || {};
        this.mode = (_b = (_a = this.options) === null || _a === void 0 ? void 0 : _a.mode) !== null && _b !== void 0 ? _b : ANCHOR_MODE.FIXED;
        this.hideOtherHandlesOnDrag = (_d = (_c = this.options) === null || _c === void 0 ? void 0 : _c.hideOtherHandlesOnDrag) !== null && _d !== void 0 ? _d : true;
        this.hideOtherControlsInstancesOnDrag = (_f = (_e = this.options) === null || _e === void 0 ? void 0 : _e.hideOtherControlsInstancesOnDrag) !== null && _f !== void 0 ? _f : true;
        this.showHelperPlane = (_h = (_g = this.options) === null || _g === void 0 ? void 0 : _g.showHelperPlane) !== null && _h !== void 0 ? _h : false;
        this.highlightAxis = (_k = (_j = this.options) === null || _j === void 0 ? void 0 : _j.highlightAxis) !== null && _k !== void 0 ? _k : true;
        this.useComputedBounds = (_m = (_l = this.options) === null || _l === void 0 ? void 0 : _l.useComputedBounds) !== null && _m !== void 0 ? _m : false;
        this.snapTranslation = (_p = (_o = this.options) === null || _o === void 0 ? void 0 : _o.snapTranslation) !== null && _p !== void 0 ? _p : {
            x: false,
            y: false,
            z: false
        };
        this.separation = (_r = (_q = this.options) === null || _q === void 0 ? void 0 : _q.separation) !== null && _r !== void 0 ? _r : DEFAULT_CONTROLS_SEPARATION;
        this.isDampingEnabled = (_t = (_s = this.options) === null || _s === void 0 ? void 0 : _s.isDampingEnabled) !== null && _t !== void 0 ? _t : true;
        this.rotationRadiusScale = (_v = (_u = this.options) === null || _u === void 0 ? void 0 : _u.rotationRadiusScale) !== null && _v !== void 0 ? _v : DEFAULT_ROTATION_RADIUS_SCALE;
        this.eyeRotationRadiusScale = (_x = (_w = this.options) === null || _w === void 0 ? void 0 : _w.eyeRotationRadiusScale) !== null && _x !== void 0 ? _x : DEFAULT_EYE_ROTATION_SCALE;
        this.pickPlaneSizeScale = (_z = (_y = this.options) === null || _y === void 0 ? void 0 : _y.pickPlaneSizeScale) !== null && _z !== void 0 ? _z : DEFAULT_PLANE_SIZE_SCALE;
        this.translationDistanceScale = (_1 = (_0 = this.options) === null || _0 === void 0 ? void 0 : _0.translationDistanceScale) !== null && _1 !== void 0 ? _1 : DEFAULT_TRANSLATION_DISTANCE_SCALE;
        if (this.options.orientation !== undefined) {
            const { x, y, z, w } = this.options.orientation;
            this.initialSelfQuaternion.set(x, y, z, w).normalize();
            this.quaternion.copy(this.initialSelfQuaternion);
        }
        this.computeObjectBounds();
        this.pick = new Pick();
        this.pickPlaneXY = new PickPlane("yellow", this.boundingSphereRadius * this.pickPlaneSizeScale, this.boundingSphereRadius * this.pickPlaneSizeScale);
        this.pickPlaneYZ = new PickPlane("cyan", this.boundingSphereRadius * this.pickPlaneSizeScale, this.boundingSphereRadius * this.pickPlaneSizeScale);
        this.pickPlaneZX = new PickPlane("pink", this.boundingSphereRadius * this.pickPlaneSizeScale, this.boundingSphereRadius * this.pickPlaneSizeScale);
        this.translationXP = new Translation("red");
        this.translationYP = new Translation("green");
        this.translationZP = new Translation("blue");
        this.translationXN = new Translation("red");
        this.translationYN = new Translation("green");
        this.translationZN = new Translation("blue");
        this.rotationX = new Rotation("red", this.boundingSphereRadius * this.rotationRadiusScale);
        this.rotationY = new Rotation("green", this.boundingSphereRadius * this.rotationRadiusScale);
        this.rotationZ = new Rotation("blue", this.boundingSphereRadius * this.rotationRadiusScale);
        this.rotationEye = new RotationEye("yellow", this.boundingSphereRadius * this.eyeRotationRadiusScale);
        this.setupDefaultTranslation();
        this.setupDefaultRotation();
        this.setupDefaultEyeRotation();
        this.setupDefaultPickPlane();
        this.setupDefaultPick();
    }
    /**
     * @hidden
     */
    getInteractiveObjects() {
        const interactiveObjects = [];
        this.handles.forEach(handle => {
            if (!handle.visible) {
                return;
            }
            interactiveObjects.push(...handle.getInteractiveObjects());
        });
        return interactiveObjects;
    }
}

function createCommonjsModule(fn) {
  var module = { exports: {} };
	return fn(module, module.exports), module.exports;
}

var nanoevents = createCommonjsModule(function (module) {
(
  /**
   * Interface for event subscription.
   *
   * @example
   * var NanoEvents = require('nanoevents')
   *
   * class Ticker {
   *   constructor() {
   *     this.emitter = new NanoEvents()
   *   }
   *   on() {
   *     return this.emitter.on.apply(this.events, arguments)
   *   }
   *   tick() {
   *     this.emitter.emit('tick')
   *   }
   * }
   *
   * @alias NanoEvents
   * @class
   */
  module.exports = function NanoEvents () {
    /**
     * Event names in keys and arrays with listeners in values.
     * @type {object}
     *
     * @example
     * Object.keys(ee.events)
     *
     * @alias NanoEvents#events
     */
    this.events = { };
  }
).prototype = {

  /**
   * Calls each of the listeners registered for a given event.
   *
   * @param {string} event The event name.
   * @param {...*} arguments The arguments for listeners.
   *
   * @return {undefined}
   *
   * @example
   * ee.emit('tick', tickType, tickDuration)
   *
   * @alias NanoEvents#emit
   * @method
   */
  emit: function emit (event) {
    var args = [].slice.call(arguments, 1)
    // Array.prototype.call() returns empty array if context is not array-like
    ;[].slice.call(this.events[event] || []).filter(function (i) {
      i.apply(null, args);
    });
  },

  /**
   * Add a listener for a given event.
   *
   * @param {string} event The event name.
   * @param {function} cb The listener function.
   *
   * @return {function} Unbind listener from event.
   *
   * @example
   * const unbind = ee.on('tick', (tickType, tickDuration) => {
   *   count += 1
   * })
   *
   * disable () {
   *   unbind()
   * }
   *
   * @alias NanoEvents#on
   * @method
   */
  on: function on (event, cb) {

    (this.events[event] = this.events[event] || []).push(cb);

    return function () {
      this.events[event] = this.events[event].filter(function (i) {
        return i !== cb
      });
    }.bind(this)
  }
};
});

/**
 * Removes all listeners.
 *
 * @param {NanoEvents} emitter NanoEvents instance.
 *
 * @returns {undefined}
 *
 * @example
 * unbindAll(emitter)
 */
function unbindAll (emitter) {
  emitter.events = { };
}

var unbindAll_1 = unbindAll;

const emitter = new nanoevents();
const unbindAll$1 = () => unbindAll_1(emitter);

const getPointFromEvent = (event) => {
    let clientX = 0;
    let clientY = 0;
    if (event instanceof MouseEvent) {
        clientX = event.clientX;
        clientY = event.clientY;
    }
    else if (event instanceof TouchEvent) {
        if (event.touches.length === 0) {
            return null;
        }
        clientX = event.touches[0].clientX;
        clientY = event.touches[0].clientY;
    }
    return { clientX, clientY };
};
const addEventListener = (element, eventNames, callback, options = false) => {
    eventNames.forEach(name => {
        element.addEventListener(name, callback, options);
    });
};
const removeEventListener = (element, eventNames, callback, options = false) => {
    eventNames.forEach(name => {
        element.removeEventListener(name, callback, options);
    });
};

var EVENTS;
(function (EVENTS) {
    EVENTS["DRAG_START"] = "DRAG_START";
    EVENTS["DRAG"] = "DRAG";
    EVENTS["DRAG_STOP"] = "DRAG_STOP";
})(EVENTS || (EVENTS = {}));
/**
 * @hidden
 * The Raycaster listens on the mouse and touch events globally and
 * dispatches DRAG_START, DRAG, and DRAG_STOP events.
 */
class Raycaster extends Raycaster$1 {
    constructor(camera, domElement, controls) {
        super();
        this.camera = camera;
        this.domElement = domElement;
        this.controls = controls;
        this.mouse = new Vector2();
        this.cameraPosition = new Vector3();
        this.activeHandle = null;
        this.activePlane = null;
        this.point = new Vector3();
        this.normal = new Vector3();
        this.visibleHandles = [];
        this.visibleControls = [];
        this.helperPlane = null;
        this.controlsWorldQuaternion = new Quaternion();
        this.clientDiagonalLength = 1;
        this.previousScreenPoint = new Vector2();
        this.currentScreenPoint = new Vector2();
        this.isActivePlaneFlipped = false;
        this.createAxisLine = () => {
            const geometry = new BufferGeometry();
            geometry.setAttribute('position', new Float32BufferAttribute([
                0, 0, -100,
                0, 0, 100
            ], 3));
            return new Line("white", geometry);
        };
        this.pointerDownListener = (event) => {
            var _a;
            const point = getPointFromEvent(event);
            // touches can be empty
            if (!point) {
                return;
            }
            const { clientX, clientY } = point;
            this.setRayDirection(clientX, clientY);
            // useful for calculating dragRatio (used in dampingFactor calculation)
            this.clientDiagonalLength = Math.sqrt(Math.pow(event.target.clientWidth, 2) +
                Math.pow(event.target.clientHeight, 2));
            this.previousScreenPoint.set(clientX, clientY);
            const interactiveObjects = [];
            Object.values(this.controls).map(controls => {
                interactiveObjects.push(...controls.getInteractiveObjects());
            });
            this.activeHandle = this.resolveHandleGroup(this.intersectObjects(interactiveObjects, true)[0]);
            if ((_a = this.activeHandle) === null || _a === void 0 ? void 0 : _a.parent) {
                const controls = this.activeHandle.parent;
                // hiding other controls and handles instances if asked
                if (controls.hideOtherControlsInstancesOnDrag) {
                    Object.values(this.controls).forEach(x => {
                        if (x.visible) {
                            this.visibleControls.push(x);
                        }
                        x.visible = false;
                    });
                    controls.visible = true;
                }
                if (controls.hideOtherHandlesOnDrag) {
                    controls.children.map(handle => {
                        if (handle.visible) {
                            this.visibleHandles.push(handle);
                        }
                        handle.visible = false;
                    });
                    this.activeHandle.visible = true;
                }
                if (this.activeHandle instanceof PickPlane) {
                    this.setPickPlaneOpacity(PICK_PLANE_OPACITY.ACTIVE);
                }
                /**
                 * creating the activePlane - the plane on which intersection actions
                 * take place. mouse movements are translated to points on the activePlane
                 */
                this.activePlane = new Plane$1();
                const eyePlaneNormal = this.getEyePlaneNormal(this.activeHandle);
                controls.getWorldQuaternion(this.controlsWorldQuaternion);
                this.normal.copy(this.activeHandle instanceof PickGroup ? eyePlaneNormal : this.activeHandle.up);
                if (!(this.activeHandle instanceof RotationEye || this.activeHandle instanceof PickGroup)) {
                    this.normal.applyQuaternion(this.controlsWorldQuaternion);
                }
                /*
                  if the angle between the eye-normal and the normal to the activePlane is
                  too small, a small mouse movement makes a large projection on the activePlane,
                  causing the object to jump big distances. To avoid this, the activePlane
                  is flipped by 90 degrees about the parallel vector of the handle.
                  This problem only requires mitigation in the TranslationGroup handle case.
                 */
                if (this.activeHandle instanceof TranslationGroup) {
                    const dot = eyePlaneNormal.dot(this.normal) / eyePlaneNormal.length();
                    // arccos(0.25) ~= 75 degrees
                    // this is the threshold to make the plane normal flip
                    this.isActivePlaneFlipped = Math.abs(dot) < 0.25;
                    if (this.isActivePlaneFlipped) {
                        this.isActivePlaneFlipped = true;
                        this.normal.applyAxisAngle(this.activeHandle.parallel, Math.PI / 2);
                    }
                }
                if (this.activeHandle instanceof TranslationGroup) {
                    this.activePlane.setFromNormalAndCoplanarPoint(this.normal, this.activeHandle.position);
                }
                else {
                    this.activePlane.setFromNormalAndCoplanarPoint(this.normal, controls.position);
                }
                // find initial intersection
                const initialIntersectionPoint = new Vector3();
                if (this.activeHandle instanceof PickGroup) {
                    this.activeHandle.getWorldPosition(initialIntersectionPoint);
                }
                else {
                    this.ray.intersectPlane(this.activePlane, initialIntersectionPoint);
                }
                // activate the helper plane if asked
                // available only for TranslationGroup and RotationGroup
                // (except RotationEye - plane of rotation is obvious)
                const scene = controls.parent;
                if (controls.showHelperPlane &&
                    (this.activeHandle instanceof TranslationGroup ||
                        this.activeHandle instanceof RotationGroup) &&
                    !(this.activeHandle instanceof RotationEye)) {
                    this.helperPlane = new PlaneHelper(this.activePlane, 1);
                    scene.add(this.helperPlane);
                }
                /**
                 * activate the highlightAxis if asked
                 * available only for TranslationGroup and RotationGroup
                 * (except RotationEye - plane of rotation is obvious)
                 */
                if (controls.highlightAxis &&
                    (this.activeHandle instanceof TranslationGroup ||
                        this.activeHandle instanceof RotationGroup) &&
                    !(this.activeHandle instanceof RotationEye)) {
                    this.activeHandle.getWorldPosition(this.highlightAxisLine.position);
                    const direction = this.highlightAxisLine.position.clone();
                    if (this.activeHandle instanceof TranslationGroup) {
                        direction.add(this.activeHandle.parallel);
                    }
                    else {
                        direction.add(this.activeHandle.up);
                    }
                    this.highlightAxisLine.lookAt(direction);
                    scene.add(this.highlightAxisLine);
                }
                // switch event listeners and dispatch DRAG_START
                removeEventListener(this.domElement, ["pointerdown", "touchstart"], this.pointerDownListener, {
                    capture: true
                });
                emitter.emit(EVENTS.DRAG_START, {
                    point: initialIntersectionPoint,
                    handle: this.activeHandle
                });
                addEventListener(this.domElement, ["pointermove", "touchmove"], this.pointerMoveListener, {
                    passive: false,
                    capture: true
                });
            }
            else {
                this.activePlane = null;
            }
        };
        this.getEyePlaneNormal = (object) => {
            this.cameraPosition.copy(this.camera.position);
            return this.cameraPosition.sub(object.position);
        };
        this.setRayDirection = (clientX, clientY) => {
            const rect = this.domElement.getBoundingClientRect();
            const { clientHeight, clientWidth } = this.domElement;
            this.mouse.x = ((clientX - rect.left) / clientWidth) * 2 - 1;
            this.mouse.y = -((clientY - rect.top) / clientHeight) * 2 + 1;
            this.setFromCamera(this.mouse, this.camera);
        };
        this.pointerMoveListener = (event) => {
            if (this.activeHandle === null || this.activePlane === null) {
                return;
            }
            const point = getPointFromEvent(event);
            if (!point) {
                return;
            }
            const { clientX, clientY } = point;
            this.setRayDirection(clientX, clientY);
            this.ray.intersectPlane(this.activePlane, this.point);
            this.currentScreenPoint.set(clientX, clientY);
            const distance = this.currentScreenPoint.distanceTo(this.previousScreenPoint);
            const dragRatio = distance / (this.clientDiagonalLength || 1);
            emitter.emit(EVENTS.DRAG, {
                point: this.point,
                handle: this.activeHandle,
                dragRatio
            });
            this.previousScreenPoint.set(clientX, clientY);
        };
        this.pointerUpListener = () => {
            var _a, _b, _c, _d;
            removeEventListener(this.domElement, ["pointermove", "touchmove"], this.pointerMoveListener, {
                capture: true
            });
            addEventListener(this.domElement, ["pointerdown", "touchstart"], this.pointerDownListener, {
                passive: false,
                capture: true
            });
            emitter.emit(EVENTS.DRAG_STOP, { point: this.point, handle: this.activeHandle });
            if (((_a = this.activeHandle) === null || _a === void 0 ? void 0 : _a.parent) &&
                this.activeHandle.parent.hideOtherControlsInstancesOnDrag) {
                this.visibleControls.forEach(controls => {
                    controls.visible = true;
                });
                this.visibleControls = [];
            }
            if (((_b = this.activeHandle) === null || _b === void 0 ? void 0 : _b.parent) &&
                this.activeHandle.parent.hideOtherHandlesOnDrag) {
                this.visibleHandles.forEach(handle => {
                    handle.visible = true;
                });
                this.visibleHandles = [];
            }
            if (this.activeHandle instanceof PickPlane) {
                this.setPickPlaneOpacity(PICK_PLANE_OPACITY.INACTIVE);
            }
            const scene = (_d = (_c = this.activeHandle) === null || _c === void 0 ? void 0 : _c.parent) === null || _d === void 0 ? void 0 : _d.parent;
            if (scene) {
                if (this.helperPlane) {
                    scene.remove(this.helperPlane);
                }
                scene.remove(this.highlightAxisLine);
            }
            this.activeHandle = null;
            this.activePlane = null;
        };
        this.resolveHandleGroup = (intersectedObject) => {
            if (intersectedObject === undefined) {
                return null;
            }
            return intersectedObject.object.parent;
        };
        this.destroy = () => {
            this.activePlane = null;
            this.activeHandle = null;
            removeEventListener(this.domElement, ["pointerdown", "touchstart"], this.pointerDownListener, {
                capture: true
            });
            removeEventListener(this.domElement, ["pointermove", "touchmove"], this.pointerMoveListener, {
                capture: true
            });
            removeEventListener(this.domElement, ["pointerup", "touchend"], this.pointerUpListener, {
                capture: true
            });
        };
        this.highlightAxisLine = this.createAxisLine();
        /**
         * mousedown and touchstart are used instead of pointerdown because
         * pointermove seems to stop firing after some a few events in chrome mobile
         * this could be because of some capture/passive setting but couldn't find
         * anything useful. using touch(*) events works.
         */
        addEventListener(this.domElement, ["pointerdown", "touchstart"], this.pointerDownListener, {
            passive: false,
            capture: true
        });
        addEventListener(this.domElement, ["pointerup", "touchend"], this.pointerUpListener, {
            passive: false,
            capture: true
        });
    }
    setPickPlaneOpacity(opacity) {
        if (!(this.activeHandle instanceof PickPlane)) {
            return;
        }
        const material = this.activeHandle.plane.material;
        if (Array.isArray(material)) {
            material.map(m => {
                m.opacity = opacity;
                m.needsUpdate = true;
            });
        }
        else {
            material.opacity = opacity;
            material.needsUpdate = true;
        }
    }
}

/**
 * The ControlsManager provides helper functions to create Controls instances
 * and link them up with a Raycaster instance (reused across multiple Controls
 * instances).
 * @noInheritDoc
 */
class ControlsManager extends Object3D {
    /**
     * @param camera - the THREE.Camera instance used in the scene
     * @param domElement - the dom element on which THREE.js renderer is attached,
     * generally available as `renderer.domElement`
     */
    constructor(camera, domElement) {
        super();
        this.camera = camera;
        this.domElement = domElement;
        this.objects = {};
        this.controls = {};
        this.eventListeners = {
            [EVENTS.DRAG_START]: [],
            [EVENTS.DRAG]: [],
            [EVENTS.DRAG_STOP]: []
        };
        this.listenToEvents = () => {
            emitter.on(EVENTS.DRAG_START, ({ point, handle }) => {
                if (handle === null) {
                    return;
                }
                const controls = handle.parent;
                if (controls === null) {
                    return;
                }
                controls.processDragStart({ point, handle });
                this.eventListeners[EVENTS.DRAG_START].map(callback => {
                    callback(controls.object, handle.name);
                });
            });
            emitter.on(EVENTS.DRAG, ({ point, handle, dragRatio }) => {
                if (handle === null) {
                    return;
                }
                const controls = handle.parent;
                if (controls === null) {
                    return;
                }
                controls.processDrag({ point, handle, dragRatio });
                this.eventListeners[EVENTS.DRAG].map(callback => {
                    callback(controls.object, handle.name);
                });
            });
            emitter.on(EVENTS.DRAG_STOP, ({ handle }) => {
                if (handle === null) {
                    return;
                }
                const controls = handle.parent;
                if (controls === null) {
                    return;
                }
                controls.processDragEnd({ handle });
                this.eventListeners[EVENTS.DRAG_STOP].map(callback => {
                    callback(controls.object, handle.name);
                });
            });
        };
        /**
         * Creates a Controls instance and attaches it to the provided THREE.js object
         *
         * @param object - the object provided by the user
         * @param options
         */
        this.anchor = (object, options) => {
            const controls = this.addControls(object, options);
            this.objects[object.id] = object;
            return controls;
        };
        /**
         * Detaches the Controls instance from the provided THREE.js object
         *
         * @param object - the object provided by the user
         * @param controls - the controls instance anchored on the object
         */
        this.detach = (object, controls) => {
            if (!this.objects.hasOwnProperty(object.id)) {
                throw new Error("object should be attached first");
            }
            this.remove(controls);
            this.dispose(controls);
            delete this.objects[object.id];
            delete this.controls[controls.id];
        };
        this.addControls = (object, options) => {
            const controls = new Controls(object, this.camera, options);
            this.controls[controls.id] = controls;
            this.add(controls);
            return controls;
        };
        /**
         * Adds an event listener. Note that there is another method `addEventListener`
         * on THREE.Object3D from which this class extends but that is specific to the
         * internals of THREE.js, but not this library
         * @param event
         * @param callback - by default the second argument is the default group name
         * for the Handle involved; for a custom handle, it is the `name` property
         * set on the handle
         */
        this.listen = (event, callback) => {
            this.eventListeners[event].push(callback);
        };
        /**
         * Removes the event listener.
         * @param event
         * @param callback
         */
        this.removeListen = (event, callback) => {
            const index = this.eventListeners[event].findIndex(x => x === callback);
            if (index !== -1) {
                this.eventListeners[event].splice(index, 1);
            }
        };
        this.dispose = (object) => {
            if (object instanceof Mesh) {
                object.geometry.dispose();
                if (Array.isArray(object.material)) {
                    object.material.map(material => material.dispose());
                }
                else {
                    object.material.dispose();
                }
            }
            while (object.children.length > 0) {
                object.children.map(child => {
                    this.dispose(child);
                    object.remove(child);
                });
            }
        };
        /**
         * Destroys all Controls instances and removes all event listeners
         */
        this.destroy = () => {
            unbindAll$1();
            const scene = this.parent;
            if (scene !== null) {
                scene.remove(this);
            }
            this.dispose(this);
            Object.values(this.controls).map(control => {
                this.dispose(control);
            });
            this.rayCaster.destroy();
            this.objects = {};
            this.controls = {};
            this.eventListeners = {
                [EVENTS.DRAG_START]: [],
                [EVENTS.DRAG]: [],
                [EVENTS.DRAG_STOP]: []
            };
        };
        this.rayCaster = new Raycaster(this.camera, this.domElement, this.controls);
        this.listenToEvents();
    }
}

export { ANCHOR_MODE, ControlsManager, DEFAULT_HANDLE_GROUP_NAME, EVENTS, Pick, PickPlane, PickPlaneGroup, Rotation, RotationGroup, Translation, TranslationGroup };
//# sourceMappingURL=three-freeform-controls.js.map
